/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 10:17
  filename: D:\Projects\Apgv5_0\ApgLib\Parser.c
   purpose: Source code file for the parser component.
  
*********************************************************************/
#include "Apg.h"
#include "Private.h"

//////////////////////////////////////////////////////////////////////////
// PARSER COMPONENT FUNCTIONS
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: one-time initialization of the parser
//              Must be called once per process before a parser component
//              can be constructed and used.
//              Translates relocatable parser data to process pointers.
//
// FUNCTION: ulong ulParserInit
//
//      ARG: 1) void* vpParserInit
//              : vpXXXInit from a valid APG parser header file XXX.h
//
//      ARG: 2) void* vpMemCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//   RETURN: true on success, false on failure
//
//////////////////////////////////////////////////////////////////////////
ulong ulParserInit(void* vpParserInit, void* vpMemCtx)
{
  ulong             ulReturn = APG_FALSE;
  APG_EXEHDR*       spExe;
  ulong             ulTest;
  ulong             ulHash;
  APG_PARSER_CTX    sCtx;
  APG_PARSER_CTX*   spCtx = &sCtx;
  APG_PARSER_INIT*  spParserInit = (APG_PARSER_INIT*)vpParserInit;
  APG_MEM_CTX*      spMemCtx = (APG_MEM_CTX*)vpMemCtx;
  DASSERT(spParserInit);
  DASSERT(spMemCtx && (spMemCtx->vpValidate = (void*)spMemCtx));
  DASSERT(sizeof(ulong) == sizeof(void*));  // some algorithms and functions assume this

  // exit if already initialized
  if(spParserInit->ulInitialized == (ulong)spParserInit)
  {
    ulReturn = APG_TRUE;
    goto APG_EXIT;
  }
  
  // set up dummy parser context for error handling
  spCtx->vpValidate = (void*)spCtx;
  spCtx->vpMemCtx = vpMemCtx;
  
  // validate the data
  spExe = (APG_EXEHDR*)spParserInit->vpExeData;
  ulTest = spExe->ulCrc32;
  spExe->ulCrc32 = 0;
  ulHash = ulCRC32((uchar*)spExe, spExe->ulLength);
  PASSERT(ulHash == ulTest);
  spExe->ulCrc32 = ulTest;

  if(!spExe->ulPPT)
  {
    // convert indexes to addresses
    ulTest = ulRuleListInit(spCtx, spExe, spParserInit->vpCallbacks);
    PASSERT(ulTest);
    ulTest = ulOpcodeInit(spCtx, spExe);
    PASSERT(ulTest);
  }
  else
  {
    // convert indexes to addresses
    ulTest = ulPPTRuleListInit(spCtx, spExe, spParserInit->vpCallbacks);
    PASSERT(ulTest);
    ulTest = ulPPTOpcodeInit(spCtx, spExe);
    PASSERT(ulTest);
  }
  
  // success
  spParserInit->ulInitialized = (ulong)spParserInit;
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: create a new APG parser component
// ulParserInit() must have previously been called in the current process.
// Parser component construction fails otherwise. The reason for failing
// instead of automatically calling ulParserInit() is to prevent the
// possibility of two threads simultaneously trying to construct components.
// ulParserInit() is not thread safe.
//
// FUNCTION: void* vpParserCtor
//
//      ARG: 1) void* vpParserInit
//              : vpXXXInit from a valid APG parser header file XXX.h
//
//      ARG: 2) void* vpMemCtx
//              : pointer to a memory context previously returned
//                from vpMemCtor() or vpMemCtorA()
//
//   RETURN: pointer to a parser component context
//           NULL on failure
//
//////////////////////////////////////////////////////////////////////////
void* vpParserCtor(void* vpParserInit, void* vpMemCtx)
{
  void*             vpReturn = NULL;
  APG_PARSER_CTX*   spCtx;
  APG_PARSER_INIT*  spParserInit = (APG_PARSER_INIT*)vpParserInit;
  APG_MEM_CTX*      spMemCtx = (APG_MEM_CTX*)vpMemCtx;
  APG_EXEHDR*       spExe;

  // validate
  if(!(spParserInit && spParserInit->ulInitialized == (ulong)spParserInit))
  {
    ulErrReport(vpMemCtx, APG_PAR_ID, __LINE__, __FILE__);
    goto APG_EXIT;
  }
  if(!(spMemCtx && spMemCtx->vpValidate == (void*)spMemCtx))
  {
    ulErrReport(vpMemCtx, APG_PAR_ID, __LINE__, __FILE__);
    goto APG_EXIT;
  }

  // allocate a context
  spCtx = (APG_PARSER_CTX*)vpMemAlloc(vpMemCtx, sizeof(APG_PARSER_CTX));
  if(!spCtx)
  {
    ulErrReport(vpMemCtx, APG_PAR_ID, __LINE__, __FILE__);
    goto APG_EXIT;
  }
  
  // initialize the context
  memset((void*)spCtx, 0, sizeof(APG_PARSER_CTX));
  spCtx->vpMemCtx = vpMemCtx;
  
  spExe = (APG_EXEHDR*)spParserInit->vpExeData;
  spCtx->vpRules = (void*)((uchar*)spExe + spExe->sRuleList.ulOffset);
  spCtx->ulRuleCount = spExe->ulRuleCount;
  spCtx->ulCharSetSize = spExe->ulCharSetSize;

  // define the virtual functions
  spCtx->ulPPT = spExe->ulPPT;

  // saves recomputing them at each node visit
  spCtx->ulEOFChar = spExe->ulEOFChar;
  spCtx->ulMaxChar = spCtx->ulMaxChar;

  // always add a AST stack
  // estimate the number of AST nodes
  spCtx->vpPtNodes = vpVecCtor(spCtx->vpMemCtx, sizeof(APG_PT_NODE), 1024);
  PASSERT(spCtx->vpPtNodes);
  spCtx->vpPtValues = vpVecCtor(spCtx->vpMemCtx, sizeof(APG_PT_VALUE), 1024);
  PASSERT(spCtx->vpPtValues);
  
  // success
  spCtx->vpValidate = (void*)spCtx;
  vpReturn = (void*)spCtx;

APG_EXIT:
  return vpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Parser component destructor
//
// FUNCTION: void vParserDtor
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vParserDtor(void* vpCtx)
{
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;

  // validate the context
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  APG_PACKRATDTOR(spCtx);

  // delete the AST
  vVecDtor(spCtx->vpPtNodes);
  vVecDtor(spCtx->vpPtValues);
  
  // invalidate & delete the parser context
  spCtx->vpValidate = NULL;
  vMemFree(spCtx->vpMemCtx, (void*)spCtx);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Performs syntax analysis on (parses) an input string.
//
// FUNCTION: ulong ulParserSyntaxAnalysis
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//      ARG: 2) ulong ulStartRule
//              : specifies the grammar rule to use as the start rule
// Zero (0) is always the same start rule used when the parser was generated.
// However, any other rule may be used for parsing an input string.
// This may be any identifier RULE_XXX_rulename where XXX.h is the header
// file of the parser and "rulename" is the name of a grammar rule.
//
//      ARG: 3) uchar* ucpSrc
//              : pointer to the first byte of the input string to parse
//
//      ARG: 4) ulong ulSrcLen
//              : number of bytes in the input string
//
//      ARG: 5) void* vpData
//              : pointer to any user-defined data - used only
//                by the call back functions
//
//      ARG: 6) ulong ulTrace
//              : if true and _APG_CFG_DEBUG is defined, will print a trace
//                of the AST traversal used in syntax analysis.
//
//   RETURN: true on success, false on failure.
//    Failure means a process failure such as "out of memory". Syntax failure
//    is indicated in the parser state, not the return value. (see ulParserState())
//
//////////////////////////////////////////////////////////////////////////
ulong ulParserSyntaxAnalysis(void* vpCtx, ulong ulStartRule, uchar* ucpSrc, ulong ulSrcLen, void* vpData, ulong ulTrace)
{
  ulong           ulReturn = APG_FALSE;
  ulong           ulState = ACTIVE;
  ulong           ulMatched;
  ulong           ulErrors;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);

  PASSERT(ucpSrc);
  PASSERT(ulStartRule < spCtx->ulRuleCount);

  // clear any previous parse data
  memset((void*)&spCtx->sStats, 0, sizeof(spCtx->sStats));
  memset((void*)&spCtx->sState, 0, sizeof(spCtx->sState));
  vVecClear(spCtx->vpPtNodes);
  vVecClear(spCtx->vpPtValues);
  spCtx->sSrc.ucpStr = NULL;
  spCtx->sSrc.ulLen = 0;
  spCtx->vpData = NULL;
  spCtx->ulTrace = APG_FALSE;
  spCtx->ulPrevNode = 0;
  spCtx->ulSeq = 0;

  // set up for this parse
  spCtx->sSrc.ucpStr = ucpSrc;
  spCtx->sSrc.ulLen = ulSrcLen;
  spCtx->vpData = vpData;
  if(ulTrace)
  {
    spCtx->ulTrace = APG_TRUE;
  }

  APG_PACKRATCTOR(spCtx);

  // run the start rule
  if(spCtx->ulPPT)
  {
    // TBD: return a binary result
    vPPTRunStartRule(spCtx, ulStartRule, &ulState, &ulMatched);
  }
  else
  {
    // TBD: return a binary result
    vRunStartRule(spCtx, ulStartRule, &ulState, &ulMatched);
  }

  // return the parser state
  spCtx->sState.ulCharactersMatched = spCtx->sStats.ulCharsMaxMatch;
  spCtx->sState.ulStringLength = spCtx->sSrc.ulLen;
  spCtx->sState.ulState = NOMATCH;
  ulErrors = ulErrCount(spCtx->vpMemCtx);
  PASSERT(ulErrors == 0);
  if(ulMatched == spCtx->sSrc.ulLen)
  {
    spCtx->sState.ulState = ulState;
    spCtx->sState.ulCharactersMatched = ulMatched;
    spCtx->sState.ulStringLength = spCtx->sSrc.ulLen;
  }
  else
  {
    spCtx->sState.ulCharactersMatched = spCtx->sStats.ulCharsMaxMatch;
    spCtx->sState.ulStringLength = spCtx->sSrc.ulLen;
    spCtx->sState.ulState = NOMATCH;
  }


  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Performs semantic analysis(syntax-directed translation)
//              on a AST generated by a call to ulParserSyntaxAnalysis().
//
// FUNCTION: ulong ulParserSemanticAnalysis
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//      ARG: 2) void* vpData
//              : pointer to any user-defined data - used only
//                by the call back functions
//
//   RETURN: true on success, false on failure
//
//////////////////////////////////////////////////////////////////////////
ulong ulParserSemanticAnalysis(void* vpCtx, void* vpData)
{
  ulong           ulReturn = APG_FALSE;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  APG_RULES*      spRuleList;
  APG_PT_NODE*    spNext;
  APG_PT_NODE*    spEnd;
  APG_PT_NODE*    spPtNodes;
  APG_PT_VALUE*   spPtValues;
  APG_PT_VALUE*   spValue;
  PFN_CALLBACK    pfnCallBack;
  ulong           ulValue, ulValueLen;
  ulong           ulCallBackReturn;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  
  spRuleList = (APG_RULES*)spCtx->vpRules;
  spPtNodes = (APG_PT_NODE*)vpVecFront(spCtx->vpPtNodes);
  PASSERT(spPtNodes);
  spPtValues = (APG_PT_VALUE*)vpVecFront(spCtx->vpPtValues);
  PASSERT(spPtValues);
  spNext = spPtNodes;
  spEnd = spNext + ulVecSize(spCtx->vpPtNodes);
  
  // traverse the AST, if any
  for(; spNext < spEnd; ++spNext)
  {
    DASSERT((spNext->ulPairedNode != APG_PT_UNDEFINED));
    spValue = spPtValues + spNext->ulValue;
    ulValue = spValue->sValue.ulOffset;
    ulValueLen = spValue->sValue.ulLen;
    if(spRuleList[spValue->ulRuleIndex].usSemantics)
    {
      // process this node
      pfnCallBack = spRuleList[spValue->ulRuleIndex].pfnCallBack;
      if(spNext->ulDirection == APG_PT_DOWN)
      {
        ulCallBackReturn = pfnCallBack(vpData, SEM_PRE, ulValue, ulValueLen);
        if(ulCallBackReturn == PFN_SKIP)
        {
          // skip branch below
          spNext = spPtNodes + spNext->ulPairedNode;
        }
        else
        {
          // reports an error for anything but OK
          PASSERT(ulCallBackReturn == PFN_OK);
        }
      }
      else
      {
        ulCallBackReturn = pfnCallBack(vpData, SEM_POST, ulValue, ulValueLen);
        // ignore return value
      }
    }
  }
  
  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: returns the parser state
//
// FUNCTION: ulong ulParserState
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//      ARG: 2) APG_PARSER_STATE* spState
//              : pointer to the user's buffer to hold the state information
//
//   RETURN: returns the parser state
//
//////////////////////////////////////////////////////////////////////////
ulong ulParserState(void* vpCtx, APG_PARSER_STATE* spState)
{
  ulong           ulReturn = APG_FALSE;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  DASSERT(spState);
  
  // return the current parser state
  memcpy((void*)spState, (void*)&spCtx->sState, sizeof(APG_PARSER_STATE));
  ulReturn = APG_TRUE;
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: returns parser statistics (if available)
//              _APG_CFG_DEBUG or _APG_CFG_STATS must be defined
//
// FUNCTION: ulong ulParserStats
//
//      ARG: 1) void* vpCtx
//              : pointer to a valid parser component context
//
//      ARG: 2) APG_PARSER_STATS* spStats
//              : pointer to the user's buffer to hold the statistics information
//
//   RETURN: returns parser statistics
//
//////////////////////////////////////////////////////////////////////////
ulong ulParserStats(void* vpCtx, APG_PARSER_STATS* spStats)
{
  ulong           ulReturn = APG_FALSE;
  APG_PARSER_CTX* spCtx = (APG_PARSER_CTX*)vpCtx;
  DASSERT(spCtx && spCtx->vpValidate == (void*)spCtx);
  DASSERT(spStats);
  
  // return the current parser state
  memcpy((void*)spStats, (void*)&spCtx->sStats, sizeof(APG_PARSER_STATS));
  ulReturn = APG_TRUE;
  
  return ulReturn;
}
