/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 13:01
  filename: D:\Projects\Apgv5_0\Generator\PPTOpcodes.c
   purpose: Generates Predictive Parsing Tables (PPT), one for each opcode.
            Also converts intermediate, PPT opcodes to production format.
  
*********************************************************************/
#include "Generator.h"

#undef _PPT_TRACE
//#define _PPT_TRACE 1

// SEST walking call back function context
#define PPT_SIZE_1 1
#define PPT_SIZE_2 2
typedef struct 
{
  void*           vpMemCtx;            // memory component for this generator
  void*           vpErrCtx;         // error component context (named same as GENCTX for GENASSERT macro)
  uchar*          ucpStrTbl;        // pointer to the string table

  // character set
  ulong           ulCharSetSize;    // number of characters in the character set
  ulong           ulCharSetUsed;    // number of used characters in the character set
  ulong           ulEOFSetSize;     // number of characters in the PPT character set (includes EOF)
  ulong*          ulpMapX;          // map of used and unused characters in the character set

  // rules
  INT_RULE_LIST*  spIntRules;       // intermediate rule list
  void*           vpRulesCtx;       // production rule list
  ulong           ulRuleCount;      // the number of rules
  ulong           ulStartRule;      // index of the start rule
  ulong           ulStartRuleCount; // indicator for production opcodes as to whether to open the rule or not

  // opcodes
  void*           vpIntOpcodesCtx;  // vector of intermediate opcodes
  void*           vpOpcodeCtx;      // vector of opcodes
  ulong           ulOpcodeCount;    // number of production opcodes
  
  // PPT
  void*           vpIntPPTCtx;      // vector context of the intermediate PPTs
  void*           vpPPTCtx;         // vector context of the PPTs
  ulong           ulPPTSize1;       // number of bytes in a 1-bit PPT
  ulong           ulPPTSize2;       // number of bytes in a 2-bit PPT
  uchar*          ucpScratchPPT;    // scratch space for PPT computations
  uchar*          ucpChildPPT;      // the PPT that the child passes up to the parent

  // SEST specific
  void*           vpWalkerCtx;      // SESTWalker context
  void*           vpStack;          // vector context of the stack of frames used in phase III
  ulong           ulMaxTerminals;   // max terminal nodes to touch before halting SESTWalk
//  PPT_ATTR        sAttr;            // attributes of the child operator returned to the parent
} PPTCTX;

typedef struct
{
  ulong ulOpId;
  ulong ulThisOpcodeIndex;
  ulong ulThisPPTIndex;
} PPTSTACKFRAME;

typedef struct 
{
  ulong ulOpID;
  ulong ulOpIndex;
} FRAME;

#ifdef _PPT_TRACE
  // define _PPT_TRACE for debug tracing of SEST traversal
  static void vPPTTraceUp(PPTCTX* spData, OP_INT_ALL* spOp, ulong ulDepth, ulong ulChildIndex, ulong ulThis, ulong ulChild);
  static void vPPTTraceDown(PPTCTX* spData, OP_INT_ALL* spOp, ulong ulDepth, ulong ulChildIndex);
  #define TRACEPPTUP(x,o,d,i,t,c) vPPTTraceUp((x),(o),(d),(i),(t),(c))
  #define TRACEPPTDOWN(x,o,d,i) vPPTTraceDown((x),(o),(d),(i))
#else
  // no code generated if _PPT_TRACE is undefined
  #define TRACEPPTUP(x,o,d,i,t,c)
  #define TRACEPPTDOWN(x,o,d,i)
#endif

//////////////////////////////////////////////////////////////////////////
// STATIC FUNCTION PROTOTYPES
//////////////////////////////////////////////////////////////////////////
static ulong ulRuleNamePPTs(GENCTX* spData, PPTCTX* spPPT);
static ulong ulOpcodesPPTs(GENCTX* spData, PPTCTX* spPPT);
static ulong ulProdOpcodesAndPPTs(GENCTX* spData, PPTCTX* spPPT, ulong ulDisplayVerbose, ulong ulDisplayPPTs);

// Phase I - SESTWalker call back functions for rule-name-only PPTs
static ulong ulPhaseOneRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneALT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOnePRD(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneREP(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneTRG(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneTBS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseOneTLS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);

// Phase II - SESTWalker call back functions for all opcode PPTs
static ulong ulPhaseTwoRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseTwoCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);

// Phase III - SESTWalker call back functions for production opcodes and PPTs
static ulong ulPhaseThreeRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);
static ulong ulPhaseThreeNonRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData);

// helpers
static uchar* ucpCatPost(PPTCTX* spData, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp);
static ulong  ulPushProdPPT(PPTCTX* spPPT, uchar* ucpPPT);
static APG_RULES* spPushProdRule(PPTCTX* spPPT, ulong* ulpReIndex, ulong ulPPTIndex);
static ulong  ulPushProdOpcodes(PPTCTX* spPPT, void* vpStack, ulong* ulpIntToProd, OP_INT_ALL* spIntOp, ulong ulCount);
static uchar* ucpInitPPT(PPTCTX* spData, PPT_ATTR* spThisAttr);
static void   vSavePPT(PPTCTX* spData, uchar* ucpDstPPT, uchar* ucpSrcPPT);
static ulong  ulHasEmpty(uchar* ucpPPT, ulong ulSize);
static ulong  ulIsAllNorA(uchar* ucpPPT, ulong ulSize);
static ulong  ulIsAllActive(uchar* ucpPPT, ulong ulSize);
static void   vChangeTo1bit(uchar* ucpPPT, ulong ulSize);
static ulong  ulPrintIntOpcodePPTs(GENCTX* spData, char* cpStrTbl, INT_RULE_LIST* spRule, ulong ulRuleCount, void* vpOpcodes, void* vpPPT, ulong ulEofSet);
static ulong  ulPrintProdOpcodePPTs(GENCTX* spData, char* cpStrTbl, APG_RULES* spRule, ulong ulRuleCount, void* vpOpcodes, void* vpPPT, ulong ulEofSet);

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: public driver for production opcodes with PPTs
//
// FUNCTION: ulong ulPPTProductionOpcodes
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulPPTProductionOpcodes(GENCTX* spData)
{
  ulong   ulReturn = APG_FALSE;
  PPTCTX  sPPTData;
  ulong   ulTest;
  char*   cpError = "ulPPTProductionOpcodes:";

  // create a production rule list vector
  GENASSERT(spData->vpRuleCtx == NULL);
  GENASSERT(spData->vpIntRuleCtx);
  spData->vpRuleCtx = vpVecCtor(spData->vpMemCtx, sizeof(APG_RULES), spData->ulRuleCount);
  GENASSERT(spData->vpRuleCtx);
  
  // create production opcode vector
  GENASSERT(spData->vpOpcodeCtx == NULL);
  GENASSERT(spData->vpIntOpcodesCtx);
  spData->vpOpcodeCtx = vpVecCtor(spData->vpMemCtx, sizeof(uchar), sizeof(OP_PPT_GEN) * ulVecSize(spData->vpIntOpcodesCtx));
  GENASSERT(spData->vpOpcodeCtx);
  
  // create PPT vectors
  GENASSERT(spData->vpPPTCtx == NULL);
  GENASSERT(spData->vpIntPPTCtx == NULL);
  spData->vpPPTCtx = vpVecCtor(spData->vpMemCtx, sizeof(uchar), 1000);
  GENASSERT(spData->vpPPTCtx);
  spData->vpIntPPTCtx = vpVecCtor(spData->vpMemCtx, sizeof(uchar), 1000);
  GENASSERT(spData->vpIntPPTCtx);
  
  // initialize the SEST walker data
  memset((void*)&sPPTData, 0, sizeof(sPPTData));
  sPPTData.vpMemCtx = spData->vpMemCtx;
  GENASSERT(spData->vpStrTblCtx);
  sPPTData.ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  GENASSERT(sPPTData.ucpStrTbl);
  
  // character set
  sPPTData.ulCharSetSize = spData->ulCharSetSize;
  sPPTData.ulCharSetUsed = spData->ulCharSetUsed;
  sPPTData.ulEOFSetSize = spData->ulEOFSetSize;
  GENASSERT(spData->vpMapXCtx);
  sPPTData.ulpMapX = (ulong*)vpVecFront(spData->vpMapXCtx);
  GENASSERT(sPPTData.ulpMapX);
  
  // rules
  GENASSERT(spData->vpIntRuleCtx);
  sPPTData.spIntRules = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
  GENASSERT(sPPTData.spIntRules);
  sPPTData.vpRulesCtx = spData->vpRuleCtx;
  sPPTData.ulRuleCount = spData->ulRuleCount;
  sPPTData.ulStartRule = spData->ulStartRule;
  
  // opcodes
  sPPTData.vpIntOpcodesCtx = spData->vpIntOpcodesCtx;
  sPPTData.vpOpcodeCtx = spData->vpOpcodeCtx;
  
  // PPTs
  sPPTData.vpIntPPTCtx = spData->vpIntPPTCtx;
  sPPTData.vpPPTCtx = spData->vpPPTCtx;
  sPPTData.ulPPTSize1 = spData->ulPPTSize1;
  sPPTData.ulPPTSize2 = spData->ulPPTSize2;
  sPPTData.ucpScratchPPT = (uchar*)vpMemAlloc(spData->vpMemCtx, 2 * spData->ulPPTSize2);
  GENASSERT(sPPTData.ucpScratchPPT);
  sPPTData.ucpChildPPT = (uchar*)vpMemAlloc(spData->vpMemCtx, 2 * spData->ulPPTSize2);
  GENASSERT(sPPTData.ucpChildPPT);
  
  // misc.
  sPPTData.ulMaxTerminals = spData->sConfig.ulMaxTerminalNodes;
  sPPTData.vpWalkerCtx = vpSESTCtor(spData, SEST_PPT, (void*)&sPPTData);
  GENASSERT(sPPTData.vpWalkerCtx);
  
  // Phase I - create PPTs for the rule names only
  ulTest = ulRuleNamePPTs(spData, &sPPTData);
  GENASSERT(ulTest);
  
  // Phase II - create PPTs for all opcodes
  ulTest = ulOpcodesPPTs(spData, &sPPTData);
  GENASSERT(ulTest);
  
  // Phase III - create production rule list, opcode and PPT vectors
  sPPTData.vpStack = vpVecCtor(spData->vpMemCtx, sizeof(PPTSTACKFRAME), 100);
  GENASSERT(sPPTData.vpStack);
  ulTest = ulProdOpcodesAndPPTs(spData, &sPPTData, spData->sConfig.ulDisplayVerbose, spData->sConfig.ulDisplayPPTs);
  GENASSERT(ulTest);
  
  // get the new opcode and rule counts
  spData->ulIntOpcodeCount = ulVecSize(spData->vpIntOpcodesCtx);
  spData->ulIntRuleCount = ulVecSize(spData->vpIntRuleCtx);
  spData->ulOpcodeCount = sPPTData.ulOpcodeCount;
  spData->ulRuleCount = sPPTData.ulRuleCount;
  
  if(spData->sConfig.ulDisplayMetrics)
  {
    // generate the metrics for the production PPTs, opcodes and SEST
    vPrintPPTGrammarMetrics(spData);
  }
  
  ulReturn = APG_TRUE;

APG_EXIT:
  // clean up unneeded resources
  if(spData->vpIntPPTCtx){vVecDtor(spData->vpIntPPTCtx); spData->vpIntPPTCtx = NULL;}
  if(spData->vpIntRuleCtx){vVecDtor(spData->vpIntRuleCtx); spData->vpIntRuleCtx = NULL;}
  if(spData->vpIntOpcodesCtx){vVecDtor(spData->vpIntOpcodesCtx); spData->vpIntOpcodesCtx = NULL;}
  if(sPPTData.ucpScratchPPT){vMemFree(spData->vpMemCtx, sPPTData.ucpScratchPPT); sPPTData.ucpScratchPPT = NULL;}
  if(sPPTData.ucpChildPPT){vMemFree(spData->vpMemCtx, sPPTData.ucpChildPPT); sPPTData.ucpChildPPT = NULL;}
  if(sPPTData.vpStack){vVecDtor(sPPTData.vpStack); sPPTData.vpStack = NULL;}
  if(sPPTData.vpWalkerCtx){vSESTDtor(sPPTData.vpWalkerCtx); sPPTData.vpWalkerCtx = NULL;}
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase I - Compute the PPTs for the rule names only.
//              This is done with a traversal of the SEST.
//              The real work is done with a set of call back functions
//              called for each node in the SEST.
//
// FUNCTION: static ulong ulRuleNamePPTs
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) PPTCTX* spPPT
//              : special context for the traversal call back functions
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulRuleNamePPTs(GENCTX* spData, PPTCTX* spPPT)
{
  ulong       ulReturn = APG_FALSE;
  ulong       ulRule;
  ulong       ulTest;
  void*       vpWalkerCtx = NULL;
  OP_INT_ALL* spOp;
  PFN_SESTCB  pfnaIntCB[OP_INT_ID_MAX];

  // single-point-of-exit loop
  // set the SEST Walker callbacks
  pfnaIntCB[OP_INT_ID_RNM] = ulPhaseOneRNM;
  pfnaIntCB[OP_INT_ID_REP] = ulPhaseOneREP;
  pfnaIntCB[OP_INT_ID_PRD] = ulPhaseOnePRD;
  pfnaIntCB[OP_INT_ID_ALT] = ulPhaseOneALT;
  pfnaIntCB[OP_INT_ID_CAT] = ulPhaseOneCAT;
  pfnaIntCB[OP_INT_ID_TRG] = ulPhaseOneTRG;
  pfnaIntCB[OP_INT_ID_TBS] = ulPhaseOneTBS;
  pfnaIntCB[OP_INT_ID_TLS] = ulPhaseOneTLS;
  vSESTSetCallBacks(spPPT->vpWalkerCtx, pfnaIntCB);
  
  // initialize the PPT indexes of all operators
  ulTest = 0;
  while(APG_TRUE)
  {
    spOp = (OP_INT_ALL*)vpVecAt(spPPT->vpIntOpcodesCtx, ulTest);
    if(!spOp){break;}
    spOp->sPPTAttr.ulIndex = UNDEFINED;
    spOp->sPPTAttr.ulComplete = APG_FALSE;
    spOp->sPPTAttr.ulTouched = APG_FALSE;
    ++ulTest;
  } // end of single-point-of-exit loop
  DASSERT(ulTest == ulVecSize(spPPT->vpIntOpcodesCtx));
  
  // initialize the PPT indexes of all rules
  ulTest = 0;
  for(ulRule = 0; ulRule < spPPT->ulRuleCount; ++ulRule)
  {
    INT_RULE_LIST* spRule = &spPPT->spIntRules[ulRule];
    spRule->sPPTAttr.ulIndex = UNDEFINED;
    spRule->sPPTAttr.ulComplete = APG_FALSE;
    spRule->sPPTAttr.ulTouched = APG_FALSE;
  } // end of single-point-of-exit loop
  
  // generate a PPT for the start rule and any rules it finds along the way
  ulTest = ulSESTWalk(spPPT->vpWalkerCtx, spPPT->ulStartRule, spPPT->ulMaxTerminals, APG_FALSE);
  GENASSERT(ulTest);
  
  // generate a PPT for all rules not already evaluated
  for(ulRule = 0; ulRule < spPPT->ulRuleCount; ++ulRule)
  {
    INT_RULE_LIST* spRule = &spPPT->spIntRules[ulRule];

    if(!spRule->sAttr.ulRef){continue;} // skip rules that are not referenced by the start rule
    if(spRule->sPPTAttr.ulComplete){continue;} // skip rules that already have PPTs
    ulTest = ulSESTWalk(spPPT->vpWalkerCtx, ulRule, spPPT->ulMaxTerminals, APG_FALSE);
    if(!ulTest){break;}
  }
  GENASSERT(ulTest);
  
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

// production opcodes after PPTs are known
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase II - Computes the production opcodes after the PPT for each
//              opcode is known.
//              This is done with a traversal of the SEST.
//              The real work is done with a set of call back functions
//              called for each node in the SEST.
//
// FUNCTION: static ulong ulOpcodesPPTs
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) PPTCTX* spPPT
//              : special context for the context functions
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulOpcodesPPTs(GENCTX* spData, PPTCTX* spPPT)
{
  ulong       ulReturn = APG_FALSE;
  ulong       ulTest;
  OP_INT_ALL* spOp;
  void*       vpWalkerCtx = NULL;
  PFN_SESTCB  pfnaIntCB[OP_INT_ID_MAX];
  
  // set the SEST Walker callbacks
  pfnaIntCB[OP_INT_ID_RNM] = ulPhaseTwoRNM;
  pfnaIntCB[OP_INT_ID_CAT] = ulPhaseTwoCAT;
  pfnaIntCB[OP_INT_ID_REP] = ulPhaseOneREP;
  pfnaIntCB[OP_INT_ID_PRD] = ulPhaseOnePRD;
  pfnaIntCB[OP_INT_ID_ALT] = ulPhaseOneALT;
  pfnaIntCB[OP_INT_ID_TRG] = ulPhaseOneTRG;
  pfnaIntCB[OP_INT_ID_TBS] = ulPhaseOneTBS;
  pfnaIntCB[OP_INT_ID_TLS] = ulPhaseOneTLS;
  vSESTSetCallBacks(spPPT->vpWalkerCtx, pfnaIntCB);
  
  // initialize the PPT indexes of all operators
  ulTest = 0;
  while(APG_TRUE)
  {
    spOp = (OP_INT_ALL*)vpVecAt(spPPT->vpIntOpcodesCtx, ulTest);
    if(!spOp){break;}
    spOp->sPPTAttr.ulIndex = UNDEFINED;
    spOp->sPPTAttr.ulComplete = APG_FALSE;
    ++ulTest;
  } // end of single-point-of-exit loop
  
  // generate a PPT for all opcodes in the SEST not already evaluated
  ulTest = ulSESTWalk(spPPT->vpWalkerCtx, spPPT->ulStartRule, spPPT->ulMaxTerminals, APG_FALSE);
  GENASSERT(ulTest);

  // success
  ulReturn = APG_TRUE;
 
APG_EXIT:
  return ulReturn;
}

// production opcodes after PPTs are known
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase II - Computes the production opcodes, weeds out
//              PPTs not used because they lie in branches below
//              determinate PPTs and will never be used.
//              This is done with a traversal of the SEST.
//              The real work is done with a set of call back functions
//              called for each node in the SEST.
//
// FUNCTION: static ulong ulProdOpcodesAndPPTs
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) PPTCTX* spPPT
//              : pointer to a special call back function context.
//
//      ARG: 3) ulong ulDisplayVerbose
//              : generator configuration display option
//
//      ARG: 4) ulong ulDisplayPPTs
//              : generator configuration display option
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulProdOpcodesAndPPTs(GENCTX* spData, PPTCTX* spPPT, ulong ulDisplayVerbose, ulong ulDisplayPPTs)
{
  ulong           ulReturn = APG_FALSE;
  ulong           ulRule;
  ulong           ulTest;
  ulong           ulProdRuleCount;
  ulong           ulProdOpcodeCount;
  uchar*          ucpPPTBeg = (uchar*)vpVecFront(spPPT->vpPPTCtx);
  APG_RULES*      spRule;
  ulong*          ulpIntToProd = NULL;
  ulong*          ulpProdToInt = NULL;
  INT_RULE_LIST*  spIntRule;
  OP_INT_ALL*     spIntOp;
  PFN_SESTCB      pfnaIntCB[OP_INT_ID_MAX];
  void*       vpStack = NULL;

    // create a temporary stack for production opcode generation
    vpStack = vpVecCtor(spPPT->vpMemCtx, sizeof(FRAME), 100);
    GENASSERT(vpStack);

    // create a temporary map (maps production rules to intermediate rules)
    ulpProdToInt = (ulong*)vpMemAlloc(spPPT->vpMemCtx, sizeof(ulong) * spPPT->ulRuleCount);
    GENASSERT(ulpProdToInt);
    // create a temporary map (maps intermediate rules to production rules)
    ulpIntToProd = (ulong*)vpMemAlloc(spPPT->vpMemCtx, sizeof(ulong) * spPPT->ulRuleCount);
    GENASSERT(ulpIntToProd);
    
    // set the SEST Walker callbacks
    pfnaIntCB[OP_INT_ID_RNM] = ulPhaseThreeRNM;
    pfnaIntCB[OP_INT_ID_REP] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_PRD] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_ALT] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_CAT] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_TRG] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_TBS] = ulPhaseThreeNonRNM;
    pfnaIntCB[OP_INT_ID_TLS] = ulPhaseThreeNonRNM;
    vSESTSetCallBacks(spPPT->vpWalkerCtx, pfnaIntCB);

    // first pass:
    // walk the start rule, tagging all rules and opcodes that will appear in the SEST
    // NOTE: some will not be tagged due to determinate PPTs canceling the branches below them
    ulTest = ulSESTWalk(spPPT->vpWalkerCtx, spPPT->ulStartRule, spPPT->ulMaxTerminals, APG_FALSE);
    GENASSERT(ulTest);
    spIntRule = spPPT->spIntRules + spPPT->ulStartRule;
    spIntRule->sPPTAttr.ulTouched = APG_TRUE;

    // map intermediate rules to production rules
    // put the start rule first
    ulProdRuleCount = 0;
    ulpProdToInt[ulProdRuleCount] = spPPT->ulStartRule;
    ulpIntToProd[spPPT->ulStartRule] = ulProdRuleCount;
    ++ulProdRuleCount;
    spIntRule = spPPT->spIntRules;
    for(ulRule = 0; ulRule < spPPT->ulRuleCount; ++ulRule, ++spIntRule)
    {
      if(ulRule == spPPT->ulStartRule){continue;}
      ulpIntToProd[ulRule] = UNDEFINED;
      if(spIntRule->sPPTAttr.ulTouched)
      {
        ulpIntToProd[ulRule] = ulProdRuleCount;
        ulpProdToInt[ulProdRuleCount] = ulRule;
        ++ulProdRuleCount;
      }
    }

    // generate production rules for each tagged rule (the ones that appear in the SEST)
    for(ulRule = 0; ulRule < ulProdRuleCount; ++ulRule)
    {
      ulong ulPPTIndex;
      uchar* ucpPPT;

      spIntRule = &spPPT->spIntRules[ulpProdToInt[ulRule]];
      DASSERT(spIntRule->sPPTAttr.ulTouched);

      // push a PPT for this rule
      ucpPPT = (uchar*)vpVecAt(spPPT->vpIntPPTCtx, spIntRule->sPPTAttr.ulIndex);
      DASSERT(ucpPPT);
      ulPPTIndex = ulPushProdPPT(spPPT, ucpPPT);
      
      // push a production rule (sets all but the opcode offset)
      spRule = spPushProdRule(spPPT, ulpProdToInt, ulPPTIndex);
      DASSERT(spRule);
    }
    GENASSERT((ulRule == ulProdRuleCount));

    // generate production opcodes for each tagged rule (the ones that appear in the SEST)
    ulProdOpcodeCount = 0;
    spRule = (APG_RULES*)vpVecFront(spPPT->vpRulesCtx);
    for(ulRule = 0; ulRule < ulProdRuleCount; ++ulRule, ++spRule)
    {
      ulong ulOpcodeOffset;

      spIntRule = &spPPT->spIntRules[ulpProdToInt[ulRule]];
      DASSERT(spIntRule->sPPTAttr.ulTouched);
      
      // process this rule into production opcodes
      ulOpcodeOffset = ulVecSize(spPPT->vpOpcodeCtx);
      spIntOp = (OP_INT_ALL*)vpVecAt(spPPT->vpIntOpcodesCtx, spIntRule->ulOpcodeOffset);
      ulTest = ulPushProdOpcodes(spPPT, vpStack, ulpIntToProd, spIntOp, spIntRule->ulOpcodeCount);
      GENASSERT((ulTest != UNDEFINED));
      ulProdOpcodeCount += ulTest;
      if(ulTest > 0)
      {
        // finish the production rule with the opcode offset
        spRule->vpOp = (OP_PPT_GEN*)ulOpcodeOffset;
      }
    }
    GENASSERT((ulRule == ulProdRuleCount));

    // reset the rule and opcode counts
    spPPT->ulOpcodeCount = ulProdOpcodeCount;
    spPPT->ulRuleCount = ulProdRuleCount;

    if(ulDisplayPPTs)
    {
      printf("\n");
      printf("*** PRODUCTION PPTS\n");
      spRule = (APG_RULES*)vpVecFront(spPPT->vpRulesCtx);
      ulTest = ulPrintProdOpcodePPTs(spData, (char*)spPPT->ucpStrTbl, spRule, spPPT->ulRuleCount, spPPT->vpOpcodeCtx, spPPT->vpPPTCtx, spPPT->ulEOFSetSize);
      DASSERT(ulTest);
    }
    
    ulReturn = APG_TRUE;

APG_EXIT:
  // release temp memory
  if(vpStack)
  {
    vVecDtor(vpStack);
    vpStack = NULL;
  }
  if(ulpProdToInt)
  {
    vMemFree(spPPT->vpMemCtx, (void*)ulpProdToInt);
    ulpProdToInt = NULL;
  }
  if(ulpIntToProd)
  {
    vMemFree(spPPT->vpMemCtx, (void*)ulpIntToProd);
    ulpIntToProd = NULL;
  }

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase I - SESTWalker call back functions for rule-name-only PPTs
//
// FUNCTION: static ulong ulPhaseOneRNM
//
//      ARG: 1) ulong ulState
//              : the traversal state SEM_PRE, SEM_POST
//
//      ARG: 2) ulong ulTreeDepth
//              : the SEST tree depth
//
//      ARG: 3) ulong ulChildCount
//              : for ALT and CAT this is the child count (1-based)
//                0 indicates before and after all children have been traversed
//
//      ARG: 4) ulong ulCountMax
//              : number of children in the child list
//
//      ARG: 5) OP_INT_ALL* spOp
//              : the opcode that this SEST tree node represents
//
//      ARG: 6) void* vpData
//              : special data context
//
//   RETURN: PFN_OK is normal return, PFN_SKIP requests caller to skip the
//           traversal of the SEST branch below this node.
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPhaseOneRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong           ulReturn = PFN_OK;
  PPTCTX*         spData = (PPTCTX*)vpData;
  INT_RULE_LIST*  spRule = &spData->spIntRules[spOp->ulRuleIndex];
  uchar*          ucpRulePPT;
  uchar*          ucpChildPPT = spData->ucpChildPPT;
  ulong           ulChar, ulValue;

  // debug validation
  DASSERT(spOp->ulOpId == OP_INT_ID_RNM);
  DASSERT(ulChildCount == ROOT_RULE);

  switch(ulState)
  {
  case SEM_PRE:
    if(spRule->sPPTAttr.ulComplete)
    {
      // do nothing - PPT already has been computed
      spOp->sPPTAttr = spRule->sPPTAttr;

      // pass this PPT up to parent
      ucpRulePPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spRule->sPPTAttr.ulIndex);
      DASSERT(ucpRulePPT);
      vSavePPT(spData, ucpChildPPT, ucpRulePPT);
      ulReturn = PFN_SKIP;
      TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    }
    else
    {
      // traverse the branch below
      TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    }
    break;
  case SEM_POST:
    // initialize a PPT for this rule
    ucpRulePPT = ucpInitPPT(spData, &spRule->sPPTAttr);

    if(spRule->usSyntax || spRule->usSemantics)
    {
      // convert child PPT to insure call back functions get called
      for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
      {
        ulValue = ulGetPPTValue(ucpChildPPT, ulChar);
        if(ulValue == EMPTY || ulValue == MATCH)
        {
          ulValue = ACTIVE;
        }
        
        vSetPPTValue(ucpRulePPT, ulChar, ulValue);
      }
      
      // pass it up to parent
      vSavePPT(spData, ucpChildPPT, ucpRulePPT);
    }
    else
    {
      // set the rule PPT same as child PPT
      vSavePPT(spData, ucpRulePPT, ucpChildPPT);
    }

    spRule->sPPTAttr.ulComplete = APG_TRUE;
    spOp->sPPTAttr = spRule->sPPTAttr;
    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spRule->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneALT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpChildPPT = spData->ucpChildPPT;
  uchar*    ucpALTPPT = spData->ucpScratchPPT;
  uchar*    ucpThisPPT;
  ulong     ulValue;
  ulong     ulChar;

  // debug validation
  DASSERT(spOp->ulOpId == OP_INT_ID_ALT);
  DASSERT(spOp->ulCount > 1);

  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    {
      switch(ulChildCount)
      {
      case 0:
        // ALT is now complete
        spOp->sPPTAttr.ulComplete = APG_TRUE;
        
        // pass it up to parent
        ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
        vSavePPT(spData, ucpChildPPT, ucpThisPPT);
      	break;
      case 1:
        // create an ALT PPT and initialize to child PPT
        ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);
        vSavePPT(spData, ucpThisPPT, ucpChildPPT);
      	break;
      default:
        // compute the new ALT PPT from old ALT PPT and child PPT
        ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
        vSavePPT(spData, ucpALTPPT, ucpThisPPT);
        
        for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
        {
          ulValue = ulGetPPTValue(ucpALTPPT, ulChar);
          if(ulValue == NOMATCH)
          {
            ulValue = ulGetPPTValue(ucpChildPPT, ulChar);
            vSetPPTValue(ucpThisPPT, ulChar, ulValue);
          }
          else
          {
            vSetPPTValue(ucpThisPPT, ulChar, ulValue);
          }
        } // end loop over characters
        break;
      }

      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    }
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;

  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_CAT);
  DASSERT(spOp->ulCount > 1);

  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    // single-point-of-exit loop
    while(APG_TRUE)
    {
      if(ulChildCount == 0)
      {
        // All CAT children have been processed - CAT is now complete
        spOp->sPPTAttr.ulComplete = APG_TRUE;

        // pass the PPT up to the parent
        ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
        vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
        break;
      }
    
      // compute CAT PPT from previous PPT and child PPT
      ucpThisPPT = ucpCatPost(spData, ulTreeDepth, ulChildCount, 0, spOp);
      DASSERT(ucpThisPPT);
      if(ulHasEmpty(ucpThisPPT, spData->ulEOFSetSize))
      {
        // CAT is not complete - continue
        spOp->sPPTAttr.ulComplete = APG_FALSE;
        break;
      }

      // NOTE: during phase I CAT must not proceed beyond the first non-empty child
      //       otherwise we risk hitting an RNM leaf for which we don't yet know the PPT attributes
      spOp->sPPTAttr.ulComplete = APG_TRUE;
      ulReturn = PFN_SKIP;
      break;
    } // end of single-point-of-exit loop

    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Computes a PPT for this CAT node. Common code for
//              Phases I and II.
//
// FUNCTION: static uchar* ucpCatPost
//
//      ARG: 1) PPTCTX* spData
//              : special data context
//
//      ARG: 2) ulong ulTreeDepth
//              : the SEST tree depth
//
//      ARG: 3) ulong ulChildCount
//              : for ALT and CAT this is the child count (1-based)
//                0 indicates before and after all children have been traversed
//
//      ARG: 4) ulong ulCountMax
//              : number of children in the child list
//
//      ARG: 5) OP_INT_ALL* spOp
//              : the opcode that this SEST tree node represents
//
//   RETURN: pointer to the new PPT
//
//////////////////////////////////////////////////////////////////////////
static uchar* ucpCatPost(PPTCTX*   spData, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp)
{
  uchar*    ucpChildPPT = spData->ucpChildPPT;
  uchar*    ucpCATPPT = spData->ucpScratchPPT;
  uchar*    ucpThisPPT = NULL;
  ulong     ulValue;
  ulong     ulChildValue;
  ulong     ulChar;

  // exit loop
  while(APG_TRUE)
  {
    DASSERT(ulChildCount != 0);

    // compute the CAT PPT from the child PPT
    if(ulChildCount == 1)
    {
      // initialize a PPT for this opcode
      ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);

      // set this CAT to child
      vSavePPT(spData, ucpThisPPT, spData->ucpChildPPT);

      // convert MATCH to ACTIVE
      for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
      {
        ulValue = ulGetPPTValue(ucpThisPPT, ulChar);
        if(ulValue == MATCH)
        {
          vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);
        }
      } // end loop over characters
      
      break;
    }

    if(ulChildCount > 1)
    {
      // compute CAT PPT from the combination of CAT and child PPTs
      ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
      vSavePPT(spData, ucpCATPPT, ucpThisPPT);
      for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
      {
        ulValue = ulGetPPTValue(ucpCATPPT, ulChar);
        switch(ulValue)
        {
        case NOMATCH:
          vSetPPTValue(ucpThisPPT, ulChar, NOMATCH);
          break;
        case EMPTY:
          ulChildValue = ulGetPPTValue(ucpChildPPT, ulChar);
          if(ulChildCount == ulCountMax)
          {
            // this is the last child in the concatenation E + M = M
            vSetPPTValue(ucpThisPPT, ulChar, ulChildValue);
          }
          else
          {
            if(ulChildValue == MATCH)
            {
              // there are more children to the right of this ont, E + M = A
              vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);
            }
            else
            {
              vSetPPTValue(ucpThisPPT, ulChar, ulChildValue);
            }
          }
          break;
        case MATCH:
        case ACTIVE:
          vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);
          break;
        default:
          DASSERT(APG_FALSE);
          break;
        }
      } // end loop over characters

      break;
    }
    break;
  } // single point of exit on multiple tests

  DASSERT(ucpThisPPT);
  return ucpThisPPT;
}

static ulong ulPhaseOnePRD(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  ulong     ulChar;
  ulong     ulValue;
  uchar*    ucpThisPPT;
  uchar*    ucpChildPPT = spData->ucpChildPPT;

  // debug validation
  DASSERT(spOp->ulOpId == OP_INT_ID_PRD);
  
  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    {
      // single-point-of-exit loop
      while(APG_TRUE)
      {
        // push and initialize a 2-bit PPT
        ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);
        
        for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
        {
          ulValue = ulGetPPTValue(ucpChildPPT, ulChar);
          switch(ulValue)
          {
          case MATCH:
          case EMPTY:
            ulValue = (spOp->ulAnd) ? EMPTY : NOMATCH;
            vSetPPTValue(ucpThisPPT, ulChar, ulValue);
            break;
          case NOMATCH:
            ulValue = (spOp->ulAnd) ? NOMATCH : EMPTY;
            vSetPPTValue(ucpThisPPT, ulChar, ulValue);
            break;
          case ACTIVE:
            vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);
            break;
          default:
            GENASSERT(APG_FALSE);
            break;
          }
        }
        
        // complete the PPT attributes
        spOp->sPPTAttr.ulComplete = APG_TRUE;
        
        // pass the PPT up to the parent
        vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
        break;
      } // end of single-point-of-exit loop
      
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    }
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneREP(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  ulong*    ulpMapX = spData->ulpMapX;
  ulong     ulChar;
  ulong     ulValue;
  uchar*    ucpThisPPT;
  uchar*    ucpChildPPT = spData->ucpChildPPT;

  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_REP);
  DASSERT(!(spOp->ulMin == 0 && spOp->ulMax == 0));

  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    {
      // push and initialize a 2-bit PPT
      ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);
      
      for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
      {
        ulValue = ulGetPPTValue(ucpChildPPT, ulChar);
        switch(ulValue)
        {
        case NOMATCH:
          ulValue = (spOp->ulMin == 0) ? EMPTY : NOMATCH;
          vSetPPTValue(ucpThisPPT, ulChar, ulValue);
          break;
        case EMPTY:
          vSetPPTValue(ucpThisPPT, ulChar, EMPTY);
          break;
        case MATCH:
          ulValue = (spOp->ulMin == 0 && spOp->ulMax == 1) ? MATCH : ACTIVE;
          vSetPPTValue(ucpThisPPT, ulChar, ulValue);
          break;
        case ACTIVE:
          vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);
          break;
        default:
          GENASSERT(APG_FALSE);
          break;
        }
      }
      
      // complete the PPT attributes
      spOp->sPPTAttr.ulComplete = APG_TRUE;
      
      // pass the PPT up to the parent
      vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    }
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneTRG(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;
  ulong     ulChar;

  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_TRG);

  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    // push and initialize a 2-bit PPT
    ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);

    // set the char flags for all characters in the range
    for(ulChar = spOp->ulMin; ulChar <= spOp->ulMax; ++ulChar)
    {
      vSetPPTValue(ucpThisPPT, ulChar, MATCH);
    }

    // complete the PPT attributes
    spOp->sPPTAttr.ulComplete = APG_TRUE;

    // pass the PPT up to the parent
    vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneTBS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;
  ulong     ulChar;
  
  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_TBS);
  DASSERT(spOp->ulLength > 1);
  
  switch(ulState)
  {
  case SEM_PRE:
    break;
  case SEM_POST:
    // push and initialize a 2-bit PPT
    ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);
    
    // set the PPT flag for the first character in the string
    ulChar = (ulong)spData->ucpStrTbl[spOp->ulOffset];
    vSetPPTValue(ucpThisPPT, ulChar, ACTIVE);

    // complete the PPT attributes
    spOp->sPPTAttr.ulComplete = APG_TRUE;
    
    // pass the PPT up to the parent
    vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseOneTLS(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;
  ulong     ulChar;
  ulong     ulMaxChar;
  ulong*    ulpMapX = (ulong*)spData->ulpMapX;
  
  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_TLS);
  
  switch(ulState)
  {
  case SEM_PRE:
    break;
  case SEM_POST:
    // push and initialize a 2-bit PPT
    ucpThisPPT = ucpInitPPT(spData, &spOp->sPPTAttr);
    
    if(spOp->ulLength == 0)
    {
      // set all used characters to EMPTY including EOF
      for(ulChar = 0; ulChar < spData->ulEOFSetSize; ++ulChar)
      {
        if(ulpMapX[ulChar] != UNDEFINED)
        {
          vSetPPTValue(ucpThisPPT, ulChar, EMPTY);
        }
      }
    }
    else
    {
      // set the PPT flag for the first character in the string
      ulong   ulCharState = (spOp->ulLength == 1) ? MATCH : ACTIVE;
      ulChar = (ulong)spData->ucpStrTbl[spOp->ulOffset];
      if(ulChar >= 65 && ulChar <= 90)
      {
        // set both upper and lower case bits
        ulMaxChar = ulChar + 32;
        vSetPPTValue(ucpThisPPT, ulMaxChar, ulCharState);
        vSetPPTValue(ucpThisPPT, ulChar, ulCharState);
      }
      if(ulChar >= 97 && ulChar <= 122)
      {
        // set both upper and lower case bits
        ulMaxChar = ulChar;
        ulChar = ulMaxChar - 32;
        vSetPPTValue(ucpThisPPT, ulMaxChar, ulCharState);
        vSetPPTValue(ucpThisPPT, ulChar, ulCharState);
      }
      else
      {
        // just set the character
        vSetPPTValue(ucpThisPPT, ulChar, ulCharState);
      }
    }

    // complete the PPT attributes
    spOp->sPPTAttr.ulComplete = APG_TRUE;
    
    // pass the PPT up to the parent
    vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase II - SESTWalker call back functions for all remaining
//              opcode PPTs not computed in Phase I.
//              All arguments same as ulPhaseOneALT
//
// FUNCTION: static ulong ulPhaseTwoRNM
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPhaseTwoRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong           ulReturn = PFN_OK;
  PPTCTX*         spData = (PPTCTX*)vpData;
  INT_RULE_LIST*  spRule = &spData->spIntRules[spOp->ulRuleIndex];
  uchar*          ucpRulePPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spRule->sPPTAttr.ulIndex);
  uchar*          ucpChildPPT = spData->ucpChildPPT;

  // debug validation
  DASSERT(spOp->ulOpId == OP_INT_ID_RNM);
  DASSERT(spRule->sPPTAttr.ulComplete);

  if(ulChildCount == ROOT_RULE)
  {
    switch(ulState)
    {
    case SEM_PRE:
      if(spRule->ulP2Complete)
      {
        // just return known rule attributes
        vSavePPT(spData, ucpChildPPT, ucpRulePPT);
        spOp->sPPTAttr = spRule->sPPTAttr;
        ulReturn = PFN_SKIP;
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
        TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
      }
      else
      {
        // open the rule and traverse the branch below
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      }
      break;
    case SEM_POST:
      // return known rule attributes
      spRule->ulP2Complete = APG_TRUE;
      vSavePPT(spData, ucpChildPPT, ucpRulePPT);
      spOp->sPPTAttr = spRule->sPPTAttr;
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spRule->sPPTAttr.ulIndex, -1);
      break;
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  } 
  else
  {
    switch(ulState)
    {
    case SEM_PRE:
      // just return known rule attributes
      vSavePPT(spData, ucpChildPPT, ucpRulePPT);
      spOp->sPPTAttr = spRule->sPPTAttr;
      ulReturn = PFN_SKIP;
      TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
      break;
    case SEM_POST:
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  }
  
APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseTwoCAT(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;

  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_CAT);
  DASSERT(spOp->ulCount > 1);

  switch(ulState)
  {
  case SEM_PRE:
    TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
    break;
  case SEM_POST:
    // single-point-of-exit loop
    while(APG_TRUE)
    {
      if(ulChildCount == 0)
      {
        // All CAT children have been processed - CAT is now complete
        spOp->sPPTAttr.ulComplete = APG_TRUE;
        
        // pass the PPT up to the parent
        ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
        vSavePPT(spData, spData->ucpChildPPT, ucpThisPPT);
        break;
      }
    
      if(!spOp->sPPTAttr.ulComplete)
      {
        // compute CAT PPT from previous PPT and child PPT
        ucpThisPPT = ucpCatPost(spData, ulTreeDepth, ulChildCount, 0, spOp);
        DASSERT(ucpThisPPT);
        if(ulHasEmpty(ucpThisPPT, spData->ulEOFSetSize))
        {
          // CAT is not complete - continue
          spOp->sPPTAttr.ulComplete = APG_FALSE;
        }
        else
        {
          // CAT is now complete
          spOp->sPPTAttr.ulComplete = APG_TRUE;
        }
      }
      break;
    } // end of single-point-of-exit loop

    TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
    break;
  default:
    GENASSERT(APG_FALSE);
    break;
  }
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// Phase III - SESTWalker call back functions for production opcodes and PPTs
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Phase III - SESTWalker call back functions for
//              production opcodes and new sparser list of PPTs.
//              NOTE: some PPTs may be eliminated because the reside
//                    in branches below a determinate PPT and will never
//                    be called upon.
//              All arguments same as ulPhaseOneALT()
//
// FUNCTION: static ulong ulPhaseThreeRNM
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPhaseThreeRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong           ulReturn = PFN_OK;
  PPTCTX*         spData = (PPTCTX*)vpData;
  INT_RULE_LIST*  spRule = &spData->spIntRules[spOp->ulRuleIndex];
  uchar*          ucpThisPPT;

  // debug validations
  DASSERT(spOp->ulOpId == OP_INT_ID_RNM);

  if(ulChildCount == ROOT_RULE)
  {
    switch(ulState)
    {
    case SEM_PRE:
      // tag the rule and opcode
      spRule->sPPTAttr.ulTouched = APG_TRUE;
      spOp->sPPTAttr.ulTouched = APG_TRUE;
      ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spRule->sPPTAttr.ulIndex);
      if(ulIsDeterminate(ucpThisPPT, spData->ulEOFSetSize))
      {
        // PPT is determinate - skip the branch below
        ulReturn = PFN_SKIP;
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
        TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, -1, -1);
      }
      else
      {
        // traverse the branch below        
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      }
      break;
    case SEM_POST:
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spRule->sPPTAttr.ulIndex, -1);
      break;
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  } 
  else
  {
    // turn around and go back up the branch
    switch(ulState)
    {
    case SEM_PRE:
      // tag the opcode
      spOp->sPPTAttr.ulTouched = APG_TRUE;
      ulReturn = PFN_SKIP;
      TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, spOp->sPPTAttr.ulIndex, -1);
      break;
    case SEM_POST:
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  }
  

APG_EXIT:
  return ulReturn;
}

static ulong ulPhaseThreeNonRNM(ulong ulState, ulong ulTreeDepth, ulong ulChildCount, ulong ulCountMax, OP_INT_ALL* spOp, void* vpData)
{
  ulong     ulReturn = PFN_OK;
  PPTCTX*   spData = (PPTCTX*)vpData;
  uchar*    ucpThisPPT;
  
  if(ulChildCount == 0)
  {
    switch(ulState)
    {
    case SEM_PRE:
      // tag the opcode
      spOp->sPPTAttr.ulTouched = APG_TRUE;
      ucpThisPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spOp->sPPTAttr.ulIndex);
      if(ulIsDeterminate(ucpThisPPT, spData->ulEOFSetSize))
      {
        // PPT is determinate - skip the branch below
        ulReturn = PFN_SKIP;
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
        TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, -1, -1);
      }
      else
      {
        // traverse the branch below        
        TRACEPPTDOWN(spData, spOp, ulTreeDepth, ulChildCount);
      }
      break;
    case SEM_POST:
      TRACEPPTUP(spData, spOp, ulTreeDepth, ulChildCount, -1, -1);
      break;
    default:
      GENASSERT(APG_FALSE);
      break;
    }
  }

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// HELPER FUNCTIONS
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Allocate and initialize a new PPT.
//
// FUNCTION: static uchar* ucpInitPPT
//
//      ARG: 1) PPTCTX* spData
//              : special context for call back functions.
//
//      ARG: 2) PPT_ATTR* spThisAttr
//              : pointer to an attribute struct for the new PPT.
//
//   RETURN: pointer to the new PPT
//
//////////////////////////////////////////////////////////////////////////
static uchar* ucpInitPPT(PPTCTX* spData, PPT_ATTR* spThisAttr)
{
  uchar* ucpThisPPT;
  
  spThisAttr->ulIndex = ulVecSize(spData->vpIntPPTCtx);
  spThisAttr->ulComplete = APG_FALSE;
  ucpThisPPT = (uchar*)vpVecPushn(spData->vpIntPPTCtx, NULL, spData->ulPPTSize2);
  memset((void*)ucpThisPPT, 0, spData->ulPPTSize2); // sets all characters to NOMATCH
  PPT_SET_TYPE_2BIT(ucpThisPPT);
  
  return ucpThisPPT;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Copy a PPT from one location to another
//
// FUNCTION: static void vSavePPT
//
//      ARG: 1) PPTCTX* spData
//              : special context for call back functions
//
//      ARG: 2) uchar* ucpDstPPT
//              : location to copy the PPT to
//
//      ARG: 3) uchar* ucpSrcPPT
//              : location to copy the PPT from
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vSavePPT(PPTCTX* spData, uchar* ucpDstPPT, uchar* ucpSrcPPT)
{
  ulong ulSize = PPT_IS_1BIT(ucpSrcPPT) ? spData->ulPPTSize1 : spData->ulPPTSize2;
  memcpy((void*)ucpDstPPT, (void*)ucpSrcPPT, ulSize);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Test a PPT. It is determinate if none of its values are ACTIVE.
//
// FUNCTION: ulong ulIsDeterminate
//
//      ARG: 1) uchar* ucpPPT
//              : pointer to the PPT
//
//      ARG: 2) ulong ulSize
//              : number of characters in the PPT character set.
//
//   RETURN: true if the PPT is determinate, false otherwise
//
//////////////////////////////////////////////////////////////////////////
ulong ulIsDeterminate(uchar* ucpPPT, ulong ulSize)
{
  ulong ulReturn = APG_TRUE;
  ulong ul;
  for(ul = 0; ul < ulSize; ++ul)
  {
    ulong ulState = ulGetPPTValue(ucpPPT, ul);
    
    if(ulState == ACTIVE)
    {
      // PPT is not determinate
      ulReturn = APG_FALSE;
      break;
    }
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Determine if the PPT has any EMPTY values.
//
// FUNCTION: static ulong ulHasEmpty
//
//      ARG: 1) uchar* ucpPPT
//              : pointer to the PPT to test
//
//      ARG: 2) ulong ulSize
//              : number of characters in the PPT character set.
//
//   RETURN: true if the PPT has at least one EMPTY value, false otherwise
//
//////////////////////////////////////////////////////////////////////////
static ulong ulHasEmpty(uchar* ucpPPT, ulong ulSize)
{
  ulong ulReturn = APG_FALSE;
  ulong ul;
  for(ul = 0; ul < ulSize; ++ul)
  {
    ulong ulState = ulGetPPTValue(ucpPPT, ul);
    
    if(ulState == EMPTY)
    {
      // PPT has at least one EMPTY character
      ulReturn = APG_TRUE;
      break;
    }
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Determine if a PPT has all values of either NOMATCH or ACTIVE.
//              (it can be reduced in size to a 1-bit representation if it does)
//
// FUNCTION: static ulong ulIsAllNorA
//
//      ARG: 1) uchar* ucpPPT
//              : pointer to the PPT to test
//
//      ARG: 2) ulong ulSize
//              : number of characters in the PPT character set.
//
//   RETURN: true if the PPT has all values of either NOMATCH or ACTIVE
//           false otherwise
//
//////////////////////////////////////////////////////////////////////////
static ulong ulIsAllNorA(uchar* ucpPPT, ulong ulSize)
{
  ulong ulReturn = APG_TRUE;
  ulong ul;
  for(ul = 0; ul < ulSize; ++ul)
  {
    ulong ulState = ulGetPPTValue(ucpPPT, ul);
    
    if(ulState == EMPTY || ulState == MATCH)
    {
      // PPT has at least one character that is not EMPTY or MATCH
      ulReturn = APG_FALSE;
      break;
    }
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Determine if a PPT has all ACTIVE values.
//              If true, the PPT is useless and can be removed.
//              NOTE: the final character in the PPT is always the special
//                    EOF character. For this test, the EOF character
//                    is ignored. This is not required, but does no damage
//                    to the accuracy of the parser. It is a specification call.
//
// FUNCTION: static ulong ulIsAllActive
//
//      ARG: 1) uchar* ucpPPT
//              : pointer to the PPT to test
//
//      ARG: 2) ulong ulSize
//              : number of characters in the PPT character set.
//
//   RETURN: true if the PPT has all values of ACTIVE
//           false otherwise
//
//////////////////////////////////////////////////////////////////////////
static ulong  ulIsAllActive(uchar* ucpPPT, ulong ulSize)
{
  ulong ulReturn = APG_TRUE;
  ulong ul;
  for(ul = 0; ul < ulSize; ++ul)
  {
    ulong ulState = ulGetPPTValue(ucpPPT, ul);
    
    if(ulState != ACTIVE)
    {
      // PPT has at least one character that is not ACTIVE
      ulReturn = APG_FALSE;
      break;
    }
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: changes a PPT in place from 2-bit to 1-bit
//
// FUNCTION: static void vChangeTo1bit
//
//      ARG: 1) uchar* ucpPPT
//              : pointer to the PPT to test
//
//      ARG: 2) ulong ulSize
//              : number of characters in the PPT character set.
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vChangeTo1bit(uchar* ucpPPT, ulong ulSize)
{
  ulong ul;
  ulong ulValue;
  for(ul = 0; ul < ulSize; ++ul)
  {
    PPT_SET_TYPE_2BIT(ucpPPT);
    ulValue = ulGetPPTValue(ucpPPT, ul);
    PPT_SET_TYPE_1BIT(ucpPPT);
    vSetPPTValue(ucpPPT, ul, ulValue);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints a human-readable list of all intermediate opcodes.
//              WARNING: This may generate a large amount of output.
//
// FUNCTION: static ulong ulPrintIntOpcodePPTs
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) char* cpStrTbl
//              : pointer to the grammar's string table
//
//      ARG: 3) INT_RULE_LIST* spRule
//              : pointer to the intermediate rule list
//
//      ARG: 4) ulong ulRuleCount
//              : number of rules in the list
//
//      ARG: 5) void* vpOpcodes
//              : pointer to the list of intermediate opcodes
//
//      ARG: 6) void* vpPPT
//              : pointer to the list of intermediate PPTs
//
//      ARG: 7) ulong ulEofSet
//              : size of the character set, including the EOF character
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPrintIntOpcodePPTs(GENCTX* spData, char* cpStrTbl, INT_RULE_LIST* spRule, ulong ulRuleCount, void* vpOpcodes, void* vpPPT, ulong ulEofSet)
{
  ulong ulReturn = APG_TRUE;
  ulong ulTest;
  char* cpName;
  OP_INT_ALL* spOpBeg;
  OP_INT_ALL* spOpEnd;
  INT_RULE_LIST* spRuleBeg = spRule;
  INT_RULE_LIST* spRuleEnd = spRule + ulRuleCount;
  
  ulTest = 0;
  for(; spRuleBeg < spRuleEnd; ++spRuleBeg)
  {
    spOpBeg = (OP_INT_ALL*)vpVecAt(vpOpcodes, spRuleBeg->ulOpcodeOffset);
    spOpEnd = spOpBeg + spRuleBeg->ulOpcodeCount;
    
    printf("\n");
    printf("rule[%s]\n", cpStrTbl + spRuleBeg->ulNameOffset);
    if(!spRule->sAttr.ulRef){continue;} // skip rules that are not referenced by the start rule
    for(; spOpBeg < spOpEnd; ++spOpBeg)
    {
      switch(spOpBeg->ulOpId)
      {
      case OP_INT_ID_ALT:
        {
          printf(" ALT(%lu)\n", spOpBeg->ulCount);
        }
        break;
      case OP_INT_ID_CAT:
        {
          printf(" CAT(%lu)\n", spOpBeg->ulCount);
        }
        break;
      case OP_INT_ID_RNM:
        {
          cpName = cpStrTbl + spRule[spOpBeg->ulRuleIndex].ulNameOffset;
          printf(" RNM(%s)\n", cpName);
        }
        break;
      case OP_INT_ID_REP:
        {
          if(spOpBeg->ulMax == UNDEFINED)
          {
            printf(" REP(%lu, inf)\n", spOpBeg->ulMin);
          } 
          else
          {
            printf(" REP(%lu, %lu)\n", spOpBeg->ulMin, spOpBeg->ulMax);
          }
        }
        break;
      case OP_INT_ID_PRD:
        {
          if(spOpBeg->ulAnd)
          {
            printf(" PRD(AND)\n");
          }
          else
          {
            printf(" PRD(NOT)\n");
          }
        }
        break;
      case OP_INT_ID_TRG:
        {
          printf(" TRG(%lu,%lu)\n", spOpBeg->ulMin, spOpBeg->ulMax);
        }
        break;
      case OP_INT_ID_TBS:
        {
          printf(" TBS(%lu)\n", spOpBeg->ulLength);
        }
        break;
      case OP_INT_ID_TLS:
        {
          printf(" TLS(%lu)\n", spOpBeg->ulLength);
        }
        break;
      default:
        GENASSERT(APG_FALSE);
        break;
      }
      if(!spOpBeg->sPPTAttr.ulComplete)
      {
        ++ulTest;
        ulReturn = APG_FALSE;
        printf(" PPT incomplete\n");
      }
      else
      {
        uchar* ucpPPT = (uchar*)vpVecAt(vpPPT, spOpBeg->sPPTAttr.ulIndex);
        vPrintPPT(ucpPPT, ulEofSet);
      }
    }
  }
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints a human-readable list of all production opcodes.
//              WARNING: This may generate a large amount of output.
//
// FUNCTION: static ulong ulPrintProdOpcodePPTs
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) char* cpStrTbl
//              : pointer to the grammar's string table
//
//      ARG: 3) INT_RULE_LIST* spRule
//              : pointer to the production rule list
//
//      ARG: 4) ulong ulRuleCount
//              : number of rules in the list
//
//      ARG: 5) void* vpOpcodes
//              : pointer to the list of production opcodes
//
//      ARG: 6) void* vpPPT
//              : pointer to the list of production PPTs
//
//      ARG: 7) ulong ulEofSet
//              : size of the character set, including the EOF character
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPrintProdOpcodePPTs(GENCTX* spData, char* cpStrTbl, APG_RULES* spRules, ulong ulRuleCount, void* vpOpcodes, void* vpPPT, ulong ulEofSet)
{
  ulong ulReturn = APG_TRUE;
  ulong ulRule;
  ulong ulNextRule;
  char* cpName;
  OP_PPT_GEN* spOpBeg;
  uchar* ucpOp;
  uchar* ucpOpEnd;
  uchar* ucpPPT;
  
  for(ulRule = 0; ulRule < ulRuleCount; ++ulRule)
  {
    printf("\n");
    printf("rule[%s]\n", cpStrTbl + (ulong)spRules[ulRule].cpRuleName);
    if((ulong)spRules[ulRule].vpOp == UNDEFINED)
    {
      if((ulong)spRules[ulRule].ucpPPT == UNDEFINED)
      {
        printf("<PPT useless - removed>\n");
      } 
      else
      {
        ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spRules[ulRule].ucpPPT);
        vPrintPPT(ucpPPT, ulEofSet);
      }
      printf("<no opcodes>\n");
      continue;
    }

    // get a pointer to the first opcode of this rule
    ucpOp = (uchar*)vpVecAt(vpOpcodes, (ulong)spRules[ulRule].vpOp);
    GENASSERT(ucpOp);

    // get a pointer to the end of this rule
    for(ulNextRule = ulRule + 1; ulNextRule <= ulRuleCount; ++ulNextRule)
    {
      if(ulNextRule == ulRuleCount)
      {
        // get the end of the opcode list
        ucpOpEnd = (uchar*)vpVecBack(vpOpcodes);
        GENASSERT(ucpOpEnd);
        ++ucpOpEnd;
      }
      else
      {
        // get the end of the opcodes for this rule
        if((ulong)spRules[ulNextRule].vpOp != UNDEFINED)
        {
          ucpOpEnd = (uchar*)vpVecAt(vpOpcodes, (ulong)spRules[ulNextRule].vpOp);
          GENASSERT(ucpOpEnd);
          break;
        }
      }
    }
    
    if((ulong)spRules[ulRule].ucpPPT == UNDEFINED)
    {
      printf("<PPT useless - removed>\n");
    } 
    else
    {
      ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spRules[ulRule].ucpPPT);
      vPrintPPT(ucpPPT, ulEofSet);
    }
    while(ucpOp < ucpOpEnd)
    {
      spOpBeg = (OP_PPT_GEN*)ucpOp;
      switch((ulong)spOpBeg->pfnOp)
      {
      case OP_INT_ID_ALT:
        {
          OP_PPT_ALT* spProdOp = (OP_PPT_ALT*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          printf(" ALT\n");
          ucpOp += sizeof(OP_PPT_ALT);

        }
        break;
      case OP_INT_ID_CAT:
        {
          OP_PPT_CAT* spProdOp = (OP_PPT_CAT*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          printf(" CAT\n");
          ucpOp += sizeof(OP_PPT_CAT);
        }
        break;
      case OP_INT_ID_RNM:
        {
          OP_PPT_RNM* spProdOp = (OP_PPT_RNM*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          cpName = cpStrTbl + (ulong)spRules[spProdOp->ulRuleIndex].cpRuleName;
          printf(" RNM(%s)\n", cpName);
          ucpOp += sizeof(OP_PPT_RNM);
        }
        break;
      case OP_INT_ID_REP:
        {
          OP_PPT_REP* spProdOp = (OP_PPT_REP*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          ucpOp += sizeof(OP_PPT_REP);
          if(spProdOp->ulMax == UNDEFINED)
          {
            printf(" REP(%lu, inf)\n", spProdOp->ulMin);
          } 
          else
          {
            printf(" REP(%lu, %lu)\n", spProdOp->ulMin, spProdOp->ulMax);
          }
        }
        break;
      case OP_INT_ID_PRD:
        {
          OP_PPT_PRD* spProdOp = (OP_PPT_PRD*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          ucpOp += sizeof(OP_PPT_PRD);
          if(spProdOp->ulAnd)
          {
            printf(" PRD(AND)\n");
          }
          else
          {
            printf(" PRD(NOT)\n");
          }
        }
        break;
      case OP_INT_ID_TRG:
        {
          OP_PPT_TRG* spProdOp = (OP_PPT_TRG*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          ucpOp += sizeof(OP_PPT_TRG);
          printf(" TRG(%lu,%lu)\n", spProdOp->ulMin, spProdOp->ulMax);
        }
        break;
      case OP_INT_ID_TBS:
        {
          OP_PPT_TBS* spProdOp = (OP_PPT_TBS*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          ucpOp += sizeof(OP_PPT_TBS);
          printf(" TBS(%lu)\n", spProdOp->ulLength);
        }
        break;
      case OP_INT_ID_TLS:
        {
          OP_PPT_TLS* spProdOp = (OP_PPT_TLS*)ucpOp;
          ucpPPT = (uchar*)vpVecAt(vpPPT, (ulong)spProdOp->ucpPPT);
          ucpOp += sizeof(OP_PPT_TLS);
          printf(" TLS(%lu)\n", spProdOp->ulLength);
        }
        break;
      default:
        GENASSERT(APG_FALSE);
        break;
      }

      if(ucpPPT == NULL)
      {
        printf("<PPT useless - removed>\n");
      } 
      else
      {
        vPrintPPT(ucpPPT, ulEofSet);
      }
    }
  }
  
APG_EXIT:
  ulReturn = (ulRule == ulRuleCount) ? APG_TRUE: APG_FALSE;
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Push a production PPT on the list of production PPTs.
//              Tests for 1-bit candidate and does the conversion if possible.
//
// FUNCTION: static ulong ulPushProdPPT
//
//      ARG: 1) PPTCTX* spData
//              : pointer to the context for the call back functions
//
//      ARG: 2) uchar* ucpPPT
//              : pointer to the PPT to push
//
//   RETURN: on success, the index of the PPT in the PPT list
//           on failure, UNDEFINED
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPushProdPPT(PPTCTX* spData, uchar* ucpPPT)
{
  ulong ulReturn = UNDEFINED;
  ulong ulIndex = ulVecSize(spData->vpPPTCtx);
  void* vpTest = NULL;
  
  if(!ulIsAllActive(ucpPPT, spData->ulCharSetSize))
  {
    if(ulIsAllNorA(ucpPPT, spData->ulEOFSetSize))
    {
      // push a 1-bit PPT
      memcpy((void*)spData->ucpScratchPPT, (void*)ucpPPT, spData->ulPPTSize2);
      vChangeTo1bit(spData->ucpScratchPPT, spData->ulEOFSetSize);
      vpTest = vpVecPushn(spData->vpPPTCtx, (void*)spData->ucpScratchPPT, spData->ulPPTSize1);
      GENASSERT(vpTest);
    }
    else
    {
      // push a 2-bit PPT
      vpTest = vpVecPushn(spData->vpPPTCtx, (void*)ucpPPT, spData->ulPPTSize2);
      GENASSERT(vpTest);
    }
  }
  // else all non-EOF PPT characters are ACTIVE - the PPT is useless - skip it

APG_EXIT:
  if(vpTest)
  {
    ulReturn = ulIndex; 
  }
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Push a production rule on the production rule list.
//
// FUNCTION: static APG_RULES* spPushProdRule
//
//      ARG: 1) PPTCTX* spPPT
//              : pointer to the context for the call back functions
//
//      ARG: 2) ulong* ulpReIndex
//              : pointer to a mapping of the intermediate rules
//                to the production rules.
//                NOTE: the rule lists may be different, since rules
//                      not referenced by the start rule are automatically
//                      removed
//
//      ARG: 3) ulong ulPPTIndex
//              : index of the PPT associated with this rule.
//
//   RETURN: pointer to the new rule
//
//////////////////////////////////////////////////////////////////////////
static APG_RULES* spPushProdRule(PPTCTX* spPPT, ulong* ulpReIndex, ulong ulPPTIndex)
{
  ulong ulRule = ulVecSize(spPPT->vpRulesCtx);
  APG_RULES* spRule = (APG_RULES*)vpVecPush(spPPT->vpRulesCtx, NULL);
  if(spRule)
  {
    INT_RULE_LIST* spIntRule = spPPT->spIntRules + ulpReIndex[ulRule];
    spRule->cpRuleName = (char*)spIntRule->ulNameOffset;  // index here - converted to pointer in ulParserInit()
    spRule->usSyntax = spIntRule->usSyntax;
    spRule->usSemantics = spIntRule->usSemantics;
    spRule->vpOp = (OP_PPT_GEN*)UNDEFINED;      // index here - converted to pointer in ulParserInit()
    spRule->ucpPPT = (uchar*)ulPPTIndex;        // index here - converted to pointer in ulParserInit()
    spRule->pfnCallBack = (PFN_CALLBACK)ulRule; // index here - converted to pointer in ulParserInit()
  }

  return spRule;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Pushes the production opcodes on the production opcode list.
//              NOTE: processing a non-terminal node requires processing
//                    the non-terminal node's children for completion.
//                    Therefore, this is function is re-entrant and
//                    called recursively. A stack of information FRAMEs
//                    is used to remember where to pick up after a child
//                    has been processed.
//
// FUNCTION: static ulong ulPushProdOpcodes
//
//      ARG: 1) PPTCTX* spData
//              : pointer to the context for the call back functions
//
//      ARG: 2) void* vpStack
//              : vector context of "FRAMES" to be used as a push/pop stack
//
//      ARG: 3) ulong* ulpIntToProd
//              : maps intermediate rule indexes to production rule indexes
//
//      ARG: 4) OP_INT_ALL* spIntOpBeg
//              : pointer to the first intermediate opcode.
//
//      ARG: 5) ulong ulCount
//              : number of intermediate opcodes in the list.
//
//   RETURN: on success, the number of opcodes in the production list
//           on failure, UNDEFINED
//
//////////////////////////////////////////////////////////////////////////
static ulong ulPushProdOpcodes(PPTCTX* spData, void* vpStack, ulong* ulpIntToProd, OP_INT_ALL* spIntOpBeg, ulong ulCount)
{
  ulong       ulReturn = UNDEFINED;
  ulong       ulOpcodeCount;
  ulong       ulProdOpcodeCount;
  ulong       ulOpcodes;
  ulong       ulTemp;
  ulong       ulTest;
  OP_INT_ALL* spIntOp;
  APG_RULES*  spRule;
  uchar*      ucpPPT;
  ulong       ulPPTIndex;
  OP_PPT_RNM* spRnm;
  OP_PPT_ALT* spAlt;
  OP_PPT_CAT* spCat;
  OP_PPT_REP* spRep;
  OP_PPT_PRD* spPrd;
  OP_PPT_TRG* spTrg;
  OP_PPT_TBS* spTbs;
  OP_PPT_TLS* spTls;
  char*       cpError = "ulPushProdOpcodes: ";
  FRAME*      spFrame;

  // single-point-of-exit loop
  ulOpcodeCount = 0;
  ulProdOpcodeCount = 0;
  while(ulOpcodeCount < ulCount)
  {
    ulTest = APG_FALSE;
    spIntOp = &spIntOpBeg[ulOpcodeCount];
    if(!spIntOp->sPPTAttr.ulTouched){++ulOpcodeCount; continue;}
    switch(spIntOp->ulOpId)
    {
    case OP_INT_ID_RNM:
      spRule = (APG_RULES*)vpVecAt(spData->vpRulesCtx, ulpIntToProd[spIntOp->ulRuleIndex]);
      spRnm = (OP_PPT_RNM*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_RNM));
      GENASSERT(spRnm);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spRnm->pfnOp = (PFN_OP)OP_INT_ID_RNM;
      spRnm->ucpPPT = spRule->ucpPPT;
      spRnm->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      spRnm->ulRuleIndex = ulpIntToProd[spIntOp->ulRuleIndex];
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_ALT:
      // push a stack frame to remember where this opcode is
      spFrame = (FRAME*)vpVecPush(vpStack, NULL);
      GENASSERT(spFrame);
      spFrame->ulOpID = OP_INT_ID_ALT;
      spFrame->ulOpIndex = ulVecSize(spData->vpOpcodeCtx);

      // push a production opcode
      spAlt = (OP_PPT_ALT*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_ALT));
      GENASSERT(spAlt);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spAlt->pfnOp = (PFN_OP)OP_INT_ID_ALT;
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);

      // push a production PPT for this opcode
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spAlt->ucpPPT = (uchar*)ulPPTIndex;

      // push production opcodes for all children
      ulOpcodes = spIntOp->ulIndexLast - spIntOp->ulIndexThis;
      ulTemp = ulPushProdOpcodes(spData, vpStack, ulpIntToProd, &spIntOp[1], ulOpcodes);
      GENASSERT((ulTemp != UNDEFINED));
      ulProdOpcodeCount += ulTemp;
      ulOpcodeCount += ulOpcodes;

      // pop the stack and finish this production opcode
      spFrame = (FRAME*)vpVecPop(vpStack);
      GENASSERT(spFrame);
      DASSERT(spFrame->ulOpID == OP_INT_ID_ALT);
      spAlt = (OP_PPT_ALT*)vpVecAt(spData->vpOpcodeCtx, spFrame->ulOpIndex);
      spAlt->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_CAT:
      // push a stack frame to remember where this opcode is
      spFrame = (FRAME*)vpVecPush(vpStack, NULL);
      GENASSERT(spFrame);
      spFrame->ulOpID = OP_INT_ID_CAT;
      spFrame->ulOpIndex = ulVecSize(spData->vpOpcodeCtx);
      
      // push a production opcode
      spCat = (OP_PPT_CAT*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_CAT));
      GENASSERT(spCat);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spCat->pfnOp = (PFN_OP)OP_INT_ID_CAT;
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      
      // push a production PPT for this opcode
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spCat->ucpPPT = (uchar*)ulPPTIndex;
      
      // push production opcodes for all children
      ulOpcodes = spIntOp->ulIndexLast - spIntOp->ulIndexThis;
      ulTemp = ulPushProdOpcodes(spData, vpStack, ulpIntToProd, &spIntOp[1], ulOpcodes);
      GENASSERT((ulTest != UNDEFINED));
      ulProdOpcodeCount += ulTemp;
      ulOpcodeCount += ulOpcodes;
      
      // pop the stack and finish this production opcode
      spFrame = (FRAME*)vpVecPop(vpStack);
      GENASSERT(spFrame);
      DASSERT(spFrame->ulOpID == OP_INT_ID_CAT);
      spCat = (OP_PPT_CAT*)vpVecAt(spData->vpOpcodeCtx, spFrame->ulOpIndex);
      spCat->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_REP:
      // push a stack frame to remember where this opcode is
      spFrame = (FRAME*)vpVecPush(vpStack, NULL);
      GENASSERT(spFrame);
      spFrame->ulOpID = OP_INT_ID_REP;
      spFrame->ulOpIndex = ulVecSize(spData->vpOpcodeCtx);
      
      // push a production opcode
      spRep = (OP_PPT_REP*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_REP));
      GENASSERT(spRep);
      spRep->pfnOp = (PFN_OP)OP_INT_ID_REP;
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      
      // push a production PPT for this opcode
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spRep->ucpPPT = (uchar*)ulPPTIndex;
      
      // push production opcodes for all children
      ulOpcodes = spIntOp->ulIndexLast - spIntOp->ulIndexThis;
      ulTemp = ulPushProdOpcodes(spData, vpStack, ulpIntToProd, &spIntOp[1], ulOpcodes);
      GENASSERT((ulTemp != UNDEFINED));
      ulProdOpcodeCount += ulTemp;
      ulOpcodeCount += ulOpcodes;
      
      // pop the stack and finish this production opcode
      spFrame = (FRAME*)vpVecPop(vpStack);
      GENASSERT(spFrame);
      DASSERT(spFrame->ulOpID == OP_INT_ID_REP);
      spRep->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      spRep->ulMin = spIntOp->ulMin;
      spRep->ulMax = spIntOp->ulMax;
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_PRD:
      spFrame = (FRAME*)vpVecPush(vpStack, NULL);
      GENASSERT(spFrame);
      spFrame->ulOpID = OP_INT_ID_PRD;
      spFrame->ulOpIndex = ulVecSize(spData->vpOpcodeCtx);
      
      // push a production opcode
      spPrd = (OP_PPT_PRD*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_PRD));
      GENASSERT(spPrd);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spPrd->pfnOp = (PFN_OP)OP_INT_ID_PRD;
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      
      // push a production PPT for this opcode
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spPrd->ucpPPT = (uchar*)ulPPTIndex;
      
      // push production opcodes for all children
      ulOpcodes = spIntOp->ulIndexLast - spIntOp->ulIndexThis;
      ulTemp = ulPushProdOpcodes(spData, vpStack, ulpIntToProd, &spIntOp[1], ulOpcodes);
      GENASSERT((ulTest != UNDEFINED));
      ulProdOpcodeCount += ulTemp;
      ulOpcodeCount += ulOpcodes;
      
      // pop the stack and finish this production opcode
      spFrame = (FRAME*)vpVecPop(vpStack);
      GENASSERT(spFrame);
      DASSERT(spFrame->ulOpID == OP_INT_ID_PRD);
      spPrd->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      spPrd->ulAnd = spIntOp->ulAnd;
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_TRG:
      spTrg = (OP_PPT_TRG*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_TRG));
      GENASSERT(spTrg);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spTrg->pfnOp = (PFN_OP)OP_INT_ID_TRG;
      spTrg->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spTrg->ucpPPT = (uchar*)ulPPTIndex;
      spTrg->ulMin = spIntOp->ulMin;
      spTrg->ulMax = spIntOp->ulMax;
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_TBS:
      spTbs = (OP_PPT_TBS*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_TBS));
      GENASSERT(spTbs);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spTbs->pfnOp = (PFN_OP)OP_INT_ID_TBS;
      spTbs->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spTbs->ucpPPT = (uchar*)ulPPTIndex;
      spTbs->ucpString = (uchar*)spIntOp->ulOffset;
      spTbs->ulLength = spIntOp->ulLength;
      ulTest = APG_TRUE;
      break;
    case OP_INT_ID_TLS:
      spTls = (OP_PPT_TLS*)vpVecPushn(spData->vpOpcodeCtx, NULL, sizeof(OP_PPT_TLS));
      GENASSERT(spTls);
      ++ulOpcodeCount;
      ++ulProdOpcodeCount;
      spTls->pfnOp = (PFN_OP)OP_INT_ID_TLS;
      spTls->spNext = (OP_PPT_GEN*)ulVecSize(spData->vpOpcodeCtx);
      ucpPPT = (uchar*)vpVecAt(spData->vpIntPPTCtx, spIntOp->sPPTAttr.ulIndex);
      GENASSERT(ucpPPT);
      ulPPTIndex = ulPushProdPPT(spData, ucpPPT);
      spTls->ucpPPT = (uchar*)ulPPTIndex;
      spTls->ucpString = (uchar*)spIntOp->ulOffset;
      spTls->ulLength = spIntOp->ulLength;
      ulTest = APG_TRUE;
      break;
    default:
      DASSERT(APG_FALSE);
      break;
    }
    GENASSERT(ulTest);
  }

APG_EXIT:
  if(ulOpcodeCount == ulCount)
  {
    ulReturn = ulProdOpcodeCount;
  }

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// TRACING FUNCTIONS
//////////////////////////////////////////////////////////////////////////
#if defined(_PPT_TRACE)
#include <stdio.h>
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints human-readable node information for a specific
//              node visit.
//
// FUNCTION: static void vTraceOp
//
//      ARG: 1) PPTCTX* spData
//              : pointer to the context for the call back functions
//
//      ARG: 2) OP_INT_ALL* spOp
//              : the opcode (node) being visited
//
//      ARG: 3) ulong ulDepth
//              : the SEST tree depth
//
//      ARG: 4) ulong ulChildIndex
//              : the child index if ALT or CAT
//
//      ARG: 5) ulong ulDown
//              : true if the traversal direction is down
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vTraceOp(PPTCTX* spData, OP_INT_ALL* spOp, ulong ulDepth, ulong ulChildIndex, ulong ulDown)
{
  char* cpName;
  INT_RULE_LIST*  spRule;
  char* cpDir = ulDown ? "down" : "up";
  
  switch(spOp->ulOpId)
  {
  case OP_INT_ID_ALT:
    printf("[->%lu] ALT(%s)[child %lu of %lu]\n", ulDepth, cpDir, ulChildIndex, spOp->ulCount);
    break;
  case OP_INT_ID_CAT:
    printf("[->%lu] CAT(%s)[child %lu of %lu]\n", ulDepth, cpDir, ulChildIndex, spOp->ulCount);
    break;
  case OP_INT_ID_PRD:
    cpName = (spOp->ulAnd) ? "AND" : "NOT";
    printf("[->%lu] PRD(%s)[%s]\n", ulDepth, cpDir, cpName);
    break;
  case OP_INT_ID_REP:
    if(spOp->ulMax == UNDEFINED)
    {
      printf("[->%lu] REP(%s)[%lu-inf]\n", ulDepth, cpDir, spOp->ulMin);
    }
    else
    {
    printf("[->%lu] REP(%s)[%lu-%lu]\n", ulDepth, cpDir, spOp->ulMin, spOp->ulMax);
    }
    break;
  case OP_INT_ID_RNM:
    spRule = &spData->spRules[spOp->ulRuleIndex];          // rule list
    cpName = (char*)spData->ucpStrTbl + spRule->ulNameOffset;
    printf("[->%lu] RNM(%s)[%s]\n", ulDepth, cpDir, cpName);
    break;
  case OP_INT_ID_TRG:
    printf("[->%lu] TRG(%s)[%lu-%lu]\n", ulDepth, cpDir, spOp->ulMin, spOp->ulMax);
    break;
  case OP_INT_ID_TBS:
    printf("[->%lu] TBS(%s)[length %lu]\n", ulDepth, cpDir, spOp->ulLength);
    break;
  case OP_INT_ID_TLS:
    printf("[->%lu] TLS(%s)[length %lu]\n", ulDepth, cpDir, spOp->ulLength);
    break;
  default:
    printf("[->%lu] ERR(%s)\n", ulDepth, cpDir);
    break;
  }

}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Called to trace a upward node visit.
//
// FUNCTION: static void vPPTTraceUp
//
//      ARG: 1) PPTCTX* spData
//              : pointer to the context for the call back functions
//
//      ARG: 2) OP_INT_ALL* spOp
//              : pointer to the opcode being visited
//
//      ARG: 3) ulong ulDepth
//              : the SEST tree depth
//
//      ARG: 4) ulong ulChildIndex
//              : the child index of ALT or CAT
//
//      ARG: 5) ulong ulThis
//              : PPT index of this opcode
//
//      ARG: 6) ulong ulChild
//              : PPT index of the child opcode
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vPPTTraceUp(PPTCTX* spData, OP_INT_ALL* spOp, ulong ulDepth, ulong ulChildIndex,
                        ulong ulThis, ulong ulChild)
{
  uchar* ucpChild = (uchar*)vpVecAt(spData->vpIntPPTCtx, ulChild);
  uchar* ucpThis = (uchar*)vpVecAt(spData->vpIntPPTCtx, ulThis);

  printf("\n");
  vTraceOp(spData, spOp, ulDepth, ulChildIndex, APG_FALSE);
  
  if(ucpChild)
  {
    printf("Child stack frame PPT\n");
    vPrintPPT(ucpChild, spData->ulEOFSetSize);
  }
  
  if(ucpThis)
  {
    printf("This stack frame PPT\n");
    vPrintPPT(ucpThis, spData->ulEOFSetSize);
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Called to trace a downward node visit.
//
// FUNCTION: static void vPPTTraceDown
//
//      ARG: 1) PPTCTX* spData
//              : pointer to the context for the call back functions
//
//      ARG: 2) OP_INT_ALL* spOp
//              : pointer to the opcode being visited
//
//      ARG: 3) ulong ulDepth
//              : the SEST tree depth
//
//      ARG: 4) ulong ulChildIndex
//              : the child index of ALT or CAT
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vPPTTraceDown(PPTCTX* spData, OP_INT_ALL* spOp, ulong ulDepth, ulong ulChildIndex)
{
  printf("\n");
  vTraceOp(spData, spOp, ulDepth, ulChildIndex, APG_TRUE);
}
#endif

